# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2018 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

"""The BIM Classification command"""

import os

import FreeCAD
import FreeCADGui

QT_TRANSLATE_NOOP = FreeCAD.Qt.QT_TRANSLATE_NOOP
translate = FreeCAD.Qt.translate

PARAMS = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/BIM")


class BIM_Classification:

    def GetResources(self):
        return {
            "Pixmap": "BIM_Classification",
            "MenuText": QT_TRANSLATE_NOOP("BIM_Classification", "Manage Classification"),
            "ToolTip": QT_TRANSLATE_NOOP(
                "BIM_Classification",
                "Manages classification systems and apply classification to objects",
            ),
        }

    def IsActive(self):
        v = hasattr(FreeCADGui.getMainWindow().getActiveWindow(), "getSceneGraph")
        return v

    def Activated(self):

        # only raise the dialog if it is already open
        if getattr(self, "form", None):
            self.form.raise_()
            return

        import Draft
        from PySide import QtCore, QtGui
        from bimcommands import BimMaterial

        # init checks
        if not hasattr(self, "Classes"):
            self.Classes = {}
        self.isEditing = None
        current = None

        # load the form and set the tree model up
        self.form = FreeCADGui.PySideUic.loadUi(":/ui/dialogClassification.ui")
        self.form.setWindowIcon(QtGui.QIcon(":/icons/BIM_Classification.svg"))
        self.form.groupMode.setItemIcon(
            0, QtGui.QIcon(":/icons/Arch_SectionPlane_Tree.svg")
        )  # Alphabetical
        self.form.groupMode.setItemIcon(1, QtGui.QIcon(":/icons/IFC.svg"))  # Type
        self.form.groupMode.setItemIcon(2, QtGui.QIcon(":/icons/Arch_Material.svg"))  # Material
        self.form.groupMode.setItemIcon(3, QtGui.QIcon(":/icons/Document.svg"))  # Model structure

        # restore saved values
        self.form.onlyVisible.setChecked(PARAMS.GetInt("BimClassificationVisibleState", 0))
        self.form.checkPrefix.setChecked(PARAMS.GetInt("BimClassificationSystemNamePrefix", 1))
        w = PARAMS.GetInt("BimClassificationDialogWidth", 629)
        h = PARAMS.GetInt("BimClassificationDialogHeight", 516)
        self.form.resize(w, h)

        # add modified search box from bimmaterial
        searchBox = BimMaterial.MatLineEdit(self.form)
        searchBox.setPlaceholderText(translate("BIM", "Search..."))
        searchBox.setToolTip(translate("BIM", "Searches classes"))
        self.form.search = searchBox
        self.form.horizontalLayout_2.addWidget(searchBox)

        # set help line
        self.form.labelDownload.setText(
            self.form.labelDownload.text().replace(
                "%s", os.path.join(FreeCAD.getUserAppDataDir(), "BIM", "Classification")
            )
        )

        # hide materials list if we are editing a particular object
        if len(FreeCADGui.Selection.getSelection()) == 1:
            self.isEditing = FreeCADGui.Selection.getSelection()[0]
            pl = self.isEditing.PropertiesList
            if ("StandardCode" in pl) or ("IfcClass" in pl):
                self.form.groupMaterials.hide()
                self.form.buttonApply.hide()
                self.form.buttonRename.hide()
                self.form.setWindowTitle(translate("BIM", "Editing") + " " + self.isEditing.Label)
                if "IfcClass" in pl:
                    # load existing class if needed
                    from nativeifc import ifc_classification

                    ifc_classification.show_classification(self.isEditing)
                if "StandardCode" in pl:
                    current = self.isEditing.StandardCode
                elif "Classification" in self.isEditing.PropertiesList:
                    current = self.isEditing.Classification

        # fill materials list
        self.objectslist = {}
        self.matlist = {}
        self.labellist = {}
        for obj in FreeCAD.ActiveDocument.Objects:
            if "StandardCode" in obj.PropertiesList:
                if Draft.getType(obj) in ["Material", "MultiMaterial"]:
                    self.matlist[obj.Name] = obj.StandardCode
                else:
                    self.objectslist[obj.Name] = obj.StandardCode
                self.labellist[obj.Name] = obj.Label
            elif "Classification" in obj.PropertiesList:
                self.objectslist[obj.Name] = obj.Classification
                self.labellist[obj.Name] = obj.Label

        # fill objects list
        if not self.isEditing:
            self.updateObjects()

        # fill available classifications
        p = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/Arch").GetString(
            "DefaultClassificationSystem", ""
        )
        presetdir = os.path.join(FreeCAD.getUserAppDataDir(), "BIM", "Classification")
        if os.path.isdir(presetdir):
            presets = []
            for f in os.listdir(presetdir):
                if f.lower().endswith(".xml") or f.lower().endswith(".ifc"):
                    n = os.path.splitext(f)[0]
                    if not n in presets:
                        presets.append(n)
                        self.form.comboSystem.addItem(n)
                        if n == p:
                            self.form.comboSystem.setCurrentIndex(self.form.comboSystem.count() - 1)

        # connect signals
        self.form.comboSystem.currentIndexChanged.connect(self.updateClasses)
        self.form.buttonApply.clicked.connect(self.apply)
        self.form.buttonRename.clicked.connect(self.rename)
        self.form.search.textEdited.connect(self.updateClasses)
        self.form.buttonBox.accepted.connect(self.accept)
        self.form.rejected.connect(self.reject)  # also triggered by self.form.buttonBox.rejected
        self.form.groupMode.currentIndexChanged.connect(self.updateObjects)
        self.form.treeClass.itemDoubleClicked.connect(self.apply)
        self.form.search.up.connect(self.onUpArrow)
        self.form.search.down.connect(self.onDownArrow)
        if hasattr(self.form.onlyVisible, "checkStateChanged"):  # Qt version >= 6.7.0
            self.form.onlyVisible.checkStateChanged.connect(self.onVisible)
            self.form.checkPrefix.checkStateChanged.connect(self.onPrefix)
        else:  # Qt version < 6.7.0
            self.form.onlyVisible.stateChanged.connect(self.onVisible)
            self.form.checkPrefix.stateChanged.connect(self.onPrefix)
        # center the dialog over FreeCAD window
        mw = FreeCADGui.getMainWindow()
        self.form.move(
            mw.frameGeometry().topLeft() + mw.rect().center() - self.form.rect().center()
        )

        self.updateClasses()

        # select current classification
        if current:
            system, classification = current.split(" ", 1)
            print("searching for", classification)
            if system in self.Classes:
                self.form.comboSystem.setCurrentText(system)
                res = self.form.treeClass.findItems(
                    classification, QtCore.Qt.MatchExactly | QtCore.Qt.MatchRecursive, 0
                )
                if res:
                    self.form.treeClass.setCurrentItem(res[0])

        self.form.show()

        self.form.search.setFocus()

    def updateObjects(self, idx=None):
        # store current state of tree into self.objectslist before redrawing

        for row in range(self.form.treeObjects.topLevelItemCount()):
            child = self.form.treeObjects.topLevelItem(row)
            if child.toolTip(0):
                if child.toolTip(0) in self.objectslist:
                    self.objectslist[child.toolTip(0)] = child.text(1)
                elif child.toolTip(0) in self.matlist:
                    self.matlist[child.toolTip(0)] = child.text(1)
                self.labellist[child.toolTip(0)] = child.text(0)
            for childrow in range(child.childCount()):
                grandchild = child.child(childrow)
                if grandchild.toolTip(0):
                    if grandchild.toolTip(0) in self.objectslist:
                        self.objectslist[grandchild.toolTip(0)] = grandchild.text(1)
                    elif grandchild.toolTip(0) in self.matlist:
                        self.matlist[grandchild.toolTip(0)] = grandchild.text(1)
                    self.labellist[grandchild.toolTip(0)] = grandchild.text(0)

        self.form.treeObjects.clear()

        if self.form.groupMode.currentIndex() == 1:
            # group by type
            self.updateByType()
        elif self.form.groupMode.currentIndex() == 2:
            # group by material
            self.updateByMaterial()
        elif self.form.groupMode.currentIndex() == 3:
            # group by model structure
            self.updateByTree()
        else:
            # group alphabetically
            self.updateDefault()

        # resize columns - no resizeSection in pyside2
        # self.form.treeObjects.header().resizeSection(0,int(self.form.treeObjects.width()/2))
        # self.form.treeObjects.header().resizeSection(1,int(self.form.treeObjects.width()/2))

    def updateByType(self):
        from PySide import QtCore, QtGui
        import Draft

        groups = {}
        for name in self.objectslist.keys():
            obj = FreeCAD.ActiveDocument.getObject(name)
            if obj and hasattr(obj, "IfcType"):
                groups.setdefault(obj.IfcType, []).append(name)
            elif obj and hasattr(obj, "IfcRole"):
                groups.setdefault(obj.IfcRole, []).append(name)
            else:
                groups.setdefault("Undefined", []).append(name)
        groups["Materials"] = self.matlist.keys()
        d = self.objectslist.copy()
        d.update(self.matlist)

        for group in groups.keys():
            mit = QtGui.QTreeWidgetItem([group, ""])
            self.form.treeObjects.addTopLevelItem(mit)
            for name in groups[group]:
                obj = FreeCAD.ActiveDocument.getObject(name)
                if obj:
                    if (
                        (not self.form.onlyVisible.isChecked())
                        or obj.ViewObject.isVisible()
                        or (Draft.getType(obj) in ["Material", "MultiMaterial"])
                    ):
                        it = QtGui.QTreeWidgetItem([self.labellist[name], d[name]])
                        it.setIcon(0, self.getIcon(obj))
                        it.setToolTip(0, name)
                        mit.addChild(it)
            mit.sortChildren(0, QtCore.Qt.AscendingOrder)
        self.form.treeObjects.expandAll()
        # self.spanTopLevels()

    def updateByMaterial(self):
        from PySide import QtCore, QtGui

        groups = {}
        claimed = []
        for name in self.matlist.keys():
            mat = FreeCAD.ActiveDocument.getObject(name)
            if mat:
                children = [par.Name for par in mat.InList if par.Name in self.objectslist.keys()]
                groups[name] = children
                claimed.extend(children)
        groups["Undefined"] = [o for o in self.objectslist.keys() if not o in claimed]

        for group in groups.keys():
            matobj = FreeCAD.ActiveDocument.getObject(group)
            if matobj:
                mit = QtGui.QTreeWidgetItem([self.labellist[group], self.matlist[group]])
                mit.setIcon(0, self.getIcon(matobj))
                mit.setToolTip(0, group)
            else:
                mit = QtGui.QTreeWidgetItem(["Undefined", ""])
            self.form.treeObjects.addTopLevelItem(mit)
            for name in groups[group]:
                obj = FreeCAD.ActiveDocument.getObject(name)
                if obj:
                    if (not self.form.onlyVisible.isChecked()) or obj.ViewObject.isVisible():
                        it = QtGui.QTreeWidgetItem([self.labellist[name], self.objectslist[name]])
                        it.setIcon(0, self.getIcon(obj))
                        it.setToolTip(0, name)
                        mit.addChild(it)
            mit.sortChildren(0, QtCore.Qt.AscendingOrder)
        self.form.treeObjects.expandAll()
        # self.spanTopLevels()

    def updateByTree(self):
        from PySide import QtGui

        # order by hierarchy
        def istop(obj):
            for parent in obj.InList:
                if parent.Name in self.objectslist.keys():
                    return False
            return True

        rel = []
        deps = []
        for name in self.objectslist.keys():
            obj = FreeCAD.ActiveDocument.getObject(name)
            if obj:
                if istop(obj):
                    rel.append(obj)
                else:
                    deps.append(obj)
        pa = 1
        while deps:
            for obj in rel:
                for child in obj.OutList:
                    if child in deps:
                        rel.append(child)
                        deps.remove(child)
            pa += 1
            if pa == 10:  # max 10 hierarchy levels, okay? Let's keep civilised
                rel.extend(deps)
                break

        done = {}
        # materials first
        mit = QtGui.QTreeWidgetItem(["Materials", ""])
        self.form.treeObjects.addTopLevelItem(mit)
        for name, code in self.matlist.items():
            obj = FreeCAD.ActiveDocument.getObject(name)
            if obj:
                it = QtGui.QTreeWidgetItem([self.labellist[name], code])
                it.setIcon(0, self.getIcon(obj))
                it.setToolTip(0, name)
                mit.addChild(it)
        # objects next
        for obj in rel:
            code = self.objectslist[obj.Name]
            if (not self.form.onlyVisible.isChecked()) or obj.ViewObject.isVisible():
                it = QtGui.QTreeWidgetItem([self.labellist[obj.Name], code])
                it.setIcon(0, self.getIcon(obj))
                it.setToolTip(0, name)
                ok = False
                for par in obj.InListRecursive:
                    if par.Name in done:
                        if (not hasattr(par, "Hosts")) or (obj not in par.Hosts):
                            done[par.Name].addChild(it)
                            done[obj.Name] = it
                            ok = True
                            break
                if not ok:
                    self.form.treeObjects.addTopLevelItem(it)
                    done[obj.Name] = it
        self.form.treeObjects.expandAll()

    def updateDefault(self):
        from PySide import QtGui
        import Draft

        d = self.objectslist.copy()
        d.update(self.matlist)
        for name, code in d.items():
            obj = FreeCAD.ActiveDocument.getObject(name)
            if obj:
                if (
                    (not self.form.onlyVisible.isChecked())
                    or obj.ViewObject.isVisible()
                    or (Draft.getType(obj) in ["Material", "MultiMaterial"])
                ):
                    it = QtGui.QTreeWidgetItem([self.labellist[name], code])
                    it.setIcon(0, self.getIcon(obj))
                    it.setToolTip(0, name)
                    self.form.treeObjects.addTopLevelItem(it)
                    if obj in FreeCADGui.Selection.getSelection():
                        self.form.treeObjects.setCurrentItem(it)

    def updateClasses(self, search=""):
        from PySide import QtGui

        self.form.treeClass.clear()

        # save as default
        FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/Arch").SetString(
            "DefaultClassificationSystem", self.form.comboSystem.currentText()
        )

        if isinstance(search, int):
            search = ""
        if self.form.search.text():
            search = self.form.search.text()
        if search:
            search = search.lower()

        system = self.form.comboSystem.currentText()
        if not system:
            return

        if not system in self.Classes:
            self.Classes[system] = self.build(system)
        if not self.Classes[system]:
            return

        for c in self.Classes[system]:
            it = None
            first = True
            if not c[1]:
                c[1] = ""
            if search:
                if (search in c[0].lower()) or (search in c[1].lower()):
                    it = QtGui.QTreeWidgetItem([c[0] + " " + c[1]])
                    it.setToolTip(0, c[1])
                    self.form.treeClass.addTopLevelItem(it)
            else:
                it = QtGui.QTreeWidgetItem([c[0] + " " + c[1]])
                it.setToolTip(0, c[1])
                self.form.treeClass.addTopLevelItem(it)
            if c[2]:
                self.addChildren(c[2], it, search)
            if it and first:
                # select first entry
                self.form.treeClass.setCurrentItem(it)
                first = False

    def addChildren(self, children, parent, search=""):
        from PySide import QtGui

        if children:
            for c in children:
                it = None
                if not c[1]:
                    c[1] = ""
                if search:
                    if (search in c[0].lower()) or (search in c[1].lower()):
                        it = QtGui.QTreeWidgetItem([c[0] + " " + c[1]])
                        it.setToolTip(0, c[1])
                        self.form.treeClass.addTopLevelItem(it)
                else:
                    it = QtGui.QTreeWidgetItem([c[0] + " " + c[1]])
                    it.setToolTip(0, c[1])
                    if parent:
                        parent.addChild(it)
                if c[2]:
                    self.addChildren(c[2], it, search)

    def build(self, system):
        # try to load the IFC first
        preset = os.path.join(FreeCAD.getUserAppDataDir(), "BIM", "Classification", system + ".ifc")
        if os.path.exists(preset):
            return self.build_ifc(system)
        else:
            preset = os.path.join(
                FreeCAD.getUserAppDataDir(), "BIM", "Classification", system + ".xml"
            )
            if os.path.exists(preset):
                return self.build_xml(system)
            else:
                FreeCAD.Console.PrintError("Unable to find classification file:" + system + "\n")
                return []

    def build_ifc(self, system):
        # builds from ifc instead of xml

        class Item:
            def __init__(self, parent=None):
                self.parent = parent
                self.ID = None
                self.Name = None
                self.children = []

        preset = os.path.join(FreeCAD.getUserAppDataDir(), "BIM", "Classification", system + ".ifc")
        if not os.path.exists(preset):
            return None
        import ifcopenshell

        f = ifcopenshell.open(preset)
        classes = f.by_type("IfcClassificationReference")
        rootclass = f.by_type("IfcClassification")
        if rootclass:
            rootclass = rootclass[0]
        else:
            return None
        root = Item()
        classdict = {rootclass.id(): root}
        for cl in classes:
            currentItem = Item()
            currentItem.Name = cl.Name
            currentItem.Description = cl.Description
            currentItem.ID = cl.Identification
            if cl.ReferencedSource:
                if cl.ReferencedSource.id() in classdict:
                    currentItem.parent = classdict[cl.ReferencedSource.id()]
                    classdict[cl.ReferencedSource.id()].children.append(currentItem)
            classdict[cl.id()] = currentItem
        return [self.listize(c) for c in root.children]

    def build_xml(self, system):
        class Item:
            def __init__(self, parent=None):
                self.parent = parent
                self.ID = None
                self.Name = None
                self.children = []

        preset = os.path.join(FreeCAD.getUserAppDataDir(), "BIM", "Classification", system + ".xml")
        if not os.path.exists(preset):
            return None
        import codecs
        import re

        d = Item()
        with codecs.open(preset, "r", "utf-8") as f:
            currentItem = d
            for l in f:
                if "<Item>" in l:
                    currentItem = Item(currentItem)
                    currentItem.parent.children.append(currentItem)
                if "</Item>" in l:
                    currentItem = currentItem.parent
                elif currentItem and re.findall(r"<ID>(.*?)</ID>", l):
                    currentItem.ID = re.findall(r"<ID>(.*?)</ID>", l)[0]
                elif currentItem and re.findall(r"<Name>(.*?)</Name>", l):
                    currentItem.Name = re.findall(r"<Name>(.*?)</Name>", l)[0]
                elif (
                    currentItem
                    and re.findall(r"<Description>(.*?)</Description>", l)
                    and not currentItem.Name
                ):
                    currentItem.Name = re.findall("<Description>(.*?)</Description>", l)[0]
        return [self.listize(c) for c in d.children]

    def listize(self, item):
        return [item.ID, item.Name, [self.listize(it) for it in item.children]]

    def apply(self, item=None, col=None):
        if self.form.treeObjects.selectedItems() and len(self.form.treeClass.selectedItems()) == 1:
            c = self.form.treeClass.selectedItems()[0].text(0)
            if self.form.checkPrefix.isChecked():
                c = self.form.comboSystem.currentText() + " " + c
            for m in self.form.treeObjects.selectedItems():
                if m.toolTip(0):
                    m.setText(1, c)

    def rename(self):
        if self.form.treeObjects.selectedItems() and len(self.form.treeClass.selectedItems()) == 1:
            c = self.form.treeClass.selectedItems()[0].toolTip(0)
            for m in self.form.treeObjects.selectedItems():
                if m.toolTip(0):
                    m.setText(0, c)

    def accept(self):
        if not self.isEditing:
            changed = False
            for row in range(self.form.treeObjects.topLevelItemCount()):
                child = self.form.treeObjects.topLevelItem(row)
                items = [child]
                items.extend([child.child(childrow) for childrow in range(child.childCount())])
                for item in items:
                    code = item.text(1)
                    label = item.text(0)
                    if item.toolTip(0):
                        obj = FreeCAD.ActiveDocument.getObject(item.toolTip(0))
                        if obj:
                            if hasattr(obj, "StandardCode"):
                                if code != obj.StandardCode:
                                    if not changed:
                                        FreeCAD.ActiveDocument.openTransaction(
                                            "Change standard codes"
                                        )
                                        changed = True
                                    obj.StandardCode = code
                            elif hasattr(obj, "IfcClass"):
                                if not "Classification" in obj.PropertiesList:
                                    obj.addProperty(
                                        "App::PropertyString", "Classification", "IFC", locked=True
                                    )
                                if code != obj.Classification:
                                    if not changed:
                                        FreeCAD.ActiveDocument.openTransaction(
                                            "Change standard codes"
                                        )
                                        changed = True
                                    obj.Classification = code
                            if label != obj.Label:
                                if not changed:
                                    FreeCAD.ActiveDocument.openTransaction("Change standard codes")
                                    changed = True
                                obj.Label = label
            if changed:
                FreeCAD.ActiveDocument.commitTransaction()
                FreeCAD.ActiveDocument.recompute()
        else:
            # Close the form if user has pressed Enter and did not
            # select anything
            if len(self.form.treeClass.selectedItems()) < 1:
                return self.reject()

            code = self.form.treeClass.selectedItems()[0].text(0)
            pl = self.isEditing.PropertiesList
            if ("StandardCode" in pl) or ("IfcClass" in pl):
                FreeCAD.ActiveDocument.openTransaction("Change standard codes")
                if self.form.checkPrefix.isChecked():
                    code = self.form.comboSystem.currentText() + " " + code
                if "StandardCode" in pl:
                    self.isEditing.StandardCode = code
                else:
                    if not "Classification" in self.isEditing.PropertiesList:
                        self.isEditing.addProperty(
                            "App::PropertyString", "Classification", "IFC", locked=True
                        )
                    self.isEditing.Classification = code
                if hasattr(self.isEditing.ViewObject, "Proxy") and hasattr(
                    self.isEditing.ViewObject.Proxy, "setTaskValue"
                ):
                    self.isEditing.ViewObject.Proxy.setTaskValue("FieldCode", code)
                FreeCAD.ActiveDocument.commitTransaction()
                FreeCAD.ActiveDocument.recompute()
        p = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/BIM")
        p.SetInt("BimClassificationDialogWidth", self.form.width())
        p.SetInt("BimClassificationDialogHeight", self.form.height())
        return self.reject()

    def reject(self):
        self.form.hide()
        del self.form
        return True

    def onUpArrow(self):
        if self.form:
            i = self.form.treeClass.currentItem()
            if self.form.treeClass.itemAbove(i):
                self.form.treeClass.setCurrentItem(self.form.treeClass.itemAbove(i))

    def onDownArrow(self):
        if self.form:
            i = self.form.treeClass.currentItem()
            if self.form.treeClass.itemBelow(i):
                self.form.treeClass.setCurrentItem(self.form.treeClass.itemBelow(i))

    def onVisible(self, index):
        PARAMS.SetInt("BimClassificationVisibleState", getattr(index, "value", index))
        self.updateObjects()

    def onPrefix(self, index):
        PARAMS.SetInt("BimClassificationSystemNamePrefix", getattr(index, "value", index))

    def getIcon(self, obj):
        """returns a QIcon for an object"""

        from PySide import QtGui
        import Arch_rc

        if hasattr(obj.ViewObject, "Icon"):
            return obj.ViewObject.Icon
        elif hasattr(obj.ViewObject, "Proxy") and hasattr(obj.ViewObject.Proxy, "getIcon"):
            icon = obj.ViewObject.Proxy.getIcon()
            if icon.startswith("/*"):
                return QtGui.QIcon(QtGui.QPixmap(icon))
            else:
                return QtGui.QIcon(icon)
        else:
            return QtGui.QIcon(":/icons/Arch_Component.svg")


FreeCADGui.addCommand("BIM_Classification", BIM_Classification())



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2024 FreeCAD Project Association                        *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

import unittest
import FreeCAD

class DocumentTestCases(unittest.TestCase):
    """
    Test class for FreeCAD material tests that need a document
    """

    def setUp(self):
        self.doc = FreeCAD.newDocument()

    def tearDown(self):
        FreeCAD.closeDocument(self.doc.Name)

    def testApplyDiffuseColorCheckShapeAppearance(self):
        """ Test that applying a DiffuseColor with transparency results in a correct ShapeAppearance """
        if "BUILD_PART" in FreeCAD.__cmake__:
            dif_col_1 = (1.0, 1.0, 0.0, 1.0)  # yellow 0% transparent
            dif_col_2 = (1.0, 0.0, 0.0, 0.5)  # red 50% transparent
            dif_col = [dif_col_1] + [dif_col_2] + 4 * [dif_col_1]

            obj = self.doc.addObject("Part::Box")
            vobj = obj.ViewObject
            vobj.DiffuseColor = dif_col

            self.assertEqual(
                [m.DiffuseColor[:3] + (1.0 - m.Transparency, ) for m in vobj.ShapeAppearance],
                vobj.DiffuseColor
            )

    def testApplyShapeAppearanceCheckDiffuseColor(self):
        """ Test that applying a ShapeAppearance with transparency results in a correct DiffuseColor """
        if "BUILD_PART" in FreeCAD.__cmake__:
            sapp_1 = FreeCAD.Material()
            sapp_1.DiffuseColor = (0.0, 1.0, 1.0, 0.0)  # cyan
            sapp_1.Transparency = 0.0                   # 0% transparent
            sapp_2 = FreeCAD.Material()
            sapp_2.DiffuseColor = (0.0, 1.0, 0.0, 0.0)  # green
            sapp_2.Transparency = 0.3                   # 30% transparent
            sapp = [sapp_1] + [sapp_2] + 4 * [sapp_1]

            obj = self.doc.addObject("Part::Box")
            vobj = obj.ViewObject
            vobj.ShapeAppearance = sapp

            self.assertEqual(
                [m.DiffuseColor[:3] + (1.0 - m.Transparency, ) for m in vobj.ShapeAppearance],
                vobj.DiffuseColor
            )



# ***************************************************************************
# *   Copyright (c) 2020 Sudhanshu Dubey <sudhanshu.thethunder@gmail.com>   *
# *   Copyright (c) 2021 Bernd Hahnebach <bernd@bimstatik.org>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import FreeCAD
from FreeCAD import Rotation
from FreeCAD import Vector

from CompoundTools import CompoundFilter

import Fem
import ObjectsFem

from . import manager
from .manager import get_meshname
from .manager import init_doc
from .meshes import generate_mesh


def get_information():
    return {
        "name": "Constraint Transform Beam Hinged",
        "meshtype": "solid",
        "meshelement": "Tet10",
        "constraints": ["pressure", "displacement", "transform"],
        "solvers": ["ccxtools"],
        "material": "solid",
        "equations": ["mechanical"],
    }


def get_explanation(header=""):
    return (
        header
        + """

To run the example from Python console use:
from femexamples.constraint_transform_beam_hinged import setup
setup()


See forum topic post:
https://forum.freecad.org/viewtopic.php?f=18&t=20238#p157643

Constraint transform on a beam

"""
    )


def setup(doc=None, solvertype="ccxtools"):

    # init FreeCAD document
    if doc is None:
        doc = init_doc()

    # explanation object
    # just keep the following line and change text string in get_explanation method
    manager.add_explanation_obj(doc, get_explanation(manager.get_header(get_information())))

    # geometric object
    # name is important because the other method in this module use obj name
    cube = doc.addObject("Part::Box", "Cube")
    cube.Height = "20 mm"
    cube.Length = "100 mm"
    cylinder = doc.addObject("Part::Cylinder", "Cylinder")
    cylinder.Height = "20 mm"
    cylinder.Radius = "6 mm"
    cylinder.Placement = FreeCAD.Placement(
        Vector(10, 12, 10),
        Rotation(0, 0, 90),
        Vector(0, 0, 0),
    )
    cut = doc.addObject("Part::Cut", "Cut")
    cut.Base = cube
    cut.Tool = cylinder

    # mirroring
    mirror = doc.addObject("Part::Mirroring", "Mirror")
    mirror.Source = cut
    mirror.Normal = (1, 0, 0)
    mirror.Base = (100, 100, 20)

    # fusing
    fusion = doc.addObject("Part::Fuse", "Fusion")
    fusion.Base = cut
    fusion.Tool = mirror
    fusion.Refine = True

    # compound filter
    geom_obj = CompoundFilter.makeCompoundFilter(name="CompoundFilter")
    geom_obj.Base = fusion
    geom_obj.FilterType = "window-volume"
    doc.recompute()

    if FreeCAD.GuiUp:
        geom_obj.Base.ViewObject.hide()
        geom_obj.ViewObject.Document.activeView().viewAxonometric()
        geom_obj.ViewObject.Document.activeView().fitAll()

    # analysis
    analysis = ObjectsFem.makeAnalysis(doc, "Analysis")

    # solver
    if solvertype == "ccxtools":
        solver_obj = ObjectsFem.makeSolverCalculiXCcxTools(doc, "CalculiXCcxTools")
        solver_obj.WorkingDir = ""
    else:
        FreeCAD.Console.PrintWarning(
            "Unknown or unsupported solver type: {}. "
            "No solver object was created.\n".format(solvertype)
        )
    if solvertype == "ccxtools":
        solver_obj.SplitInputWriter = False
        solver_obj.AnalysisType = "static"
        solver_obj.GeometricalNonlinearity = "linear"
        solver_obj.ThermoMechSteadyState = False
        solver_obj.MatrixSolverType = "default"
        solver_obj.IterationsControlParameterTimeUse = False
    analysis.addObject(solver_obj)

    # material
    material_obj = ObjectsFem.makeMaterialSolid(doc, "FemMaterial")
    mat = material_obj.Material
    mat["Name"] = "CalculiX-Steel"
    mat["YoungsModulus"] = "210000 MPa"
    mat["PoissonRatio"] = "0.30"
    material_obj.Material = mat
    analysis.addObject(material_obj)

    # constraint pressure
    con_pressure = ObjectsFem.makeConstraintPressure(doc, name="FemConstraintPressure")
    con_pressure.References = [(geom_obj, "Face8")]
    con_pressure.Pressure = "10.0 MPa"
    con_pressure.Reversed = False
    analysis.addObject(con_pressure)

    # constraint displacement
    con_disp = ObjectsFem.makeConstraintDisplacement(doc, name="FemConstraintDisplacement")
    con_disp.References = [(geom_obj, "Face4"), (geom_obj, "Face5")]
    con_disp.xFree = False
    con_disp.xDisplacement = 0.0
    analysis.addObject(con_disp)

    # constraints transform
    con_transform1 = ObjectsFem.makeConstraintTransform(doc, name="FemConstraintTransform1")
    con_transform1.References = [(geom_obj, "Face4")]
    con_transform1.TransformType = "Cylindrical"
    analysis.addObject(con_transform1)

    con_transform2 = ObjectsFem.makeConstraintTransform(doc, name="FemConstraintTransform2")
    con_transform2.References = [(geom_obj, "Face5")]
    con_transform2.TransformType = "Cylindrical"
    analysis.addObject(con_transform2)

    # mesh
    from .meshes.mesh_transform_beam_hinged_tetra10 import create_nodes, create_elements

    fem_mesh = generate_mesh.mesh_from_existing(create_nodes, create_elements)
    femmesh_obj = analysis.addObject(ObjectsFem.makeMeshGmsh(doc, get_meshname()))[0]
    femmesh_obj.FemMesh = fem_mesh
    femmesh_obj.Shape = geom_obj
    femmesh_obj.SecondOrderLinear = False
    femmesh_obj.CharacteristicLengthMax = "7 mm"

    doc.recompute()
    return doc



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2020 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************


__title__ = "FreeCAD structural IFC export tools"
__author__ = "Yorik van Havre"
__url__ = "https://www.freecad.org"

ALLOW_LINEAR_OBJECTS = True  # allow non-solid objects (wires, etc) to become analytic objects?

structural_nodes = {}  # this keeps track of nodes during this session
structural_curves = {}  # this keeps track of structural curves during this session
scaling = 1.0  # this keeps track of scaling during this session


def setup(ifcfile, ifcbin, scale):
    """Creates all the needed setup for structural model."""

    global structural_nodes, scaling
    structural_nodes = {}
    scaling = scale
    import ifcopenshell

    uid = ifcopenshell.guid.new
    ownerHistory = ifcfile.by_type("IfcOwnerHistory")[0]
    project = ifcfile.by_type("IfcProject")[0]
    structContext = createStructuralContext(ifcfile)
    if ifcfile.wrapped_data.schema_name() == "IFC2X3":
        mod = ifcfile.createIfcStructuralAnalysisModel(
            uid(),
            ownerHistory,
            "Structural Analysis Model",
            None,
            None,
            "NOTDEFINED",
            None,
            None,
            None,
        )
    else:
        localPlacement = ifcbin.createIfcLocalPlacement()
        structModel = ifcfile.createIfcStructuralAnalysisModel(
            uid(),
            ownerHistory,
            "Structural Analysis Model",
            None,
            None,
            "NOTDEFINED",
            None,
            None,
            None,
            localPlacement,
        )
        relation = ifcfile.createIfcRelDeclares(
            uid(), ownerHistory, None, None, project, [structModel]
        )


def createStructuralContext(ifcfile):
    """Creates an additional geometry context for structural objects. Returns the new context"""

    contexts = ifcfile.by_type("IfcGeometricRepresentationContext")
    # filter out subcontexts
    contexts = [c for c in contexts if c.is_a() == "IfcGeometricRepresentationContext"]
    geomContext = contexts[0]  # arbitrarily take the first one...
    structContext = ifcfile.createIfcGeometricRepresentationSubContext(
        "Analysis", "Axis", None, None, None, None, geomContext, None, "GRAPH_VIEW", None
    )
    return structContext


def getStructuralContext(ifcfile):
    """Returns the structural context from the file"""
    for c in ifcfile.by_type("IfcGeometricRepresentationSubContext"):
        if c.ContextIdentifier == "Analysis":
            return c


def createStructuralNode(ifcfile, ifcbin, point):
    """Creates a connection node at the given point"""

    import ifcopenshell

    uid = ifcopenshell.guid.new
    ownerHistory = ifcfile.by_type("IfcOwnerHistory")[0]
    structContext = getStructuralContext(ifcfile)
    cartPoint = ifcbin.createIfcCartesianPoint(tuple(point))
    vertPoint = ifcfile.createIfcVertexPoint(cartPoint)
    topologyRep = ifcfile.createIfcTopologyRepresentation(
        structContext, "Analysis", "Vertex", [vertPoint]
    )
    prodDefShape = ifcfile.createIfcProductDefinitionShape(None, None, [topologyRep])
    # boundary conditions serve for ex. to create fixed nodes
    # appliedCondition = ifcfile.createIfcBoundaryNodeCondition(
    #     "Fixed",ifcfile.createIfcBoolean(True), ifcfile.createIfcBoolean(True), ifcfile.createIfcBoolean(True),
    #     ifcfile.createIfcBoolean(True), ifcfile.createIfcBoolean(True), ifcfile.createIfcBoolean(True))
    # for now we don't create any boundary condition
    appliedCondition = None
    localPlacement = ifcbin.createIfcLocalPlacement()
    if ifcfile.wrapped_data.schema_name() == "IFC2X3":
        structPntConn = ifcfile.createIfcStructuralPointConnection(
            uid(),
            ownerHistory,
            "Vertex",
            None,
            None,
            localPlacement,
            prodDefShape,
            appliedCondition,
        )
    else:
        structPntConn = ifcfile.createIfcStructuralPointConnection(
            uid(),
            ownerHistory,
            "Vertex",
            None,
            None,
            localPlacement,
            prodDefShape,
            appliedCondition,
            None,
        )
    return structPntConn


def createStructuralCurve(ifcfile, ifcbin, curve):
    """Creates a structural connection for a curve"""

    import ifcopenshell

    uid = ifcopenshell.guid.new
    ownerHistory = ifcfile.by_type("IfcOwnerHistory")[0]
    structContext = getStructuralContext(ifcfile)

    cartPnt1 = ifcbin.createIfcCartesianPoint(tuple(curve.Vertexes[0].Point.multiply(scaling)))
    cartPnt2 = ifcbin.createIfcCartesianPoint(tuple(curve.Vertexes[-1].Point.multiply(scaling)))
    vertPnt1 = ifcfile.createIfcVertexPoint(cartPnt1)
    vertPnt2 = ifcfile.createIfcVertexPoint(cartPnt2)
    edge = ifcfile.createIfcEdge(vertPnt1, vertPnt2)
    topologyRep = ifcfile.createIfcTopologyRepresentation(structContext, "Analysis", "Edge", [edge])
    prodDefShape = ifcfile.createIfcProductDefinitionShape(None, None, [topologyRep])

    # boundary conditions serve for ex. to create fixed edges
    # for now we don't create any boundary condition
    appliedCondition = None
    localPlacement = ifcbin.createIfcLocalPlacement()
    origin = ifcfile.createIfcCartesianPoint((0.0, 0.0, 0.0))
    orientation = [[1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]]
    xAxis = ifcfile.createIfcDirection(tuple(orientation[0]))
    zAxis = ifcfile.createIfcDirection(tuple(orientation[2]))
    localAxes = ifcfile.createIfcAxis2Placement3D(origin, zAxis, xAxis)
    structCrvConn = ifcfile.createIfcStructuralCurveConnection(
        uid(),
        ownerHistory,
        "Line",
        None,
        None,
        localPlacement,
        prodDefShape,
        appliedCondition,
        localAxes,
    )
    return structCrvConn


def createStructuralMember(ifcfile, ifcbin, obj):
    """Creates a structural member if possible. Returns the member"""

    global structural_nodes, structural_curves
    structuralMember = None
    import Draft
    import Part
    import ifcopenshell
    import FreeCAD

    uid = ifcopenshell.guid.new
    ownerHistory = ifcfile.by_type("IfcOwnerHistory")[0]
    structContext = getStructuralContext(ifcfile)

    # find edges to convert into structural members
    edges = None
    if Draft.getType(obj) not in ["Structure"]:
        # for non structural elements
        if ALLOW_LINEAR_OBJECTS and obj.isDerivedFrom("Part::Feature"):
            # for objects created with Part workbench
            if obj.Shape.Faces:
                # for objects with faces
                return None
            elif not obj.Shape.Edges:
                # for objects without edges
                return None
            else:
                edges = obj.Shape.Edges
    else:
        # for structural elements with nodes
        nodes = [obj.Placement.multVec(n) for n in obj.Nodes]
        if len(nodes) > 2:
            # when there are more then 2 nodes (i.e. for slabs) append closing node to produce closing edge
            nodes.append(nodes[0])
        wire = Part.makePolygon(nodes)
        edges = wire.Edges
    if not edges:
        return None

    # OBJECT CLASSIFICATION by edge number
    # Linear elements for edge_number = 1, Surface elements for edge_number > 1
    # we don't care about curved edges just now...

    if len(edges) == 1:
        # LINEAR OBJECTS: beams, columns
        # ATM limitations:
        # - no profile properties are taken into account
        # - no materials properties are takein into account
        # -
        # create geometry
        verts = [None for _ in range(len(edges) + 1)]
        verts[0] = tuple(edges[0].Vertexes[0].Point.multiply(scaling))
        verts[1] = tuple(edges[0].Vertexes[-1].Point.multiply(scaling))
        cartPnt1 = ifcfile.createIfcCartesianPoint(verts[0])
        cartPnt2 = ifcfile.createIfcCartesianPoint(verts[1])
        vertPnt1 = ifcfile.createIfcVertexPoint(cartPnt1)
        vertPnt2 = ifcfile.createIfcVertexPoint(cartPnt2)
        newEdge = ifcfile.createIfcEdge(vertPnt1, vertPnt2)
        topologyRep = ifcfile.createIfcTopologyRepresentation(
            structContext, "Analysis", "Edge", (newEdge,)
        )
        prodDefShape = ifcfile.createIfcProductDefinitionShape(None, None, (topologyRep,))
        # set local coordinate system
        localPlacement = ifcbin.createIfcLocalPlacement()
        localZAxis = ifcbin.createIfcDirection((0, 0, 1))
        # create structural member
        if ifcfile.wrapped_data.schema_name() == "IFC2X3":
            structuralMember = ifcfile.createIfcStructuralCurveMember(
                uid(),
                ownerHistory,
                obj.Label,
                None,
                None,
                localPlacement,
                prodDefShape,
                "RIGID_JOINED_MEMBER",
            )
        else:
            localZAxis = ifcbin.createIfcDirection((0, 0, 1))
            structuralMember = ifcfile.createIfcStructuralCurveMember(
                uid(),
                ownerHistory,
                obj.Label,
                None,
                None,
                localPlacement,
                prodDefShape,
                "RIGID_JOINED_MEMBER",
                localZAxis,
            )

    elif len(edges) > 1:
        # SURFACE OBJECTS: slabs (horizontal, vertical, inclined)
        # ATM limitations:
        # - mo material properties are taken into account
        # - walls don't work because they miss a node system
        # -
        # creates geometry
        verts = [None for _ in range(len(edges))]
        for i, edge in enumerate(edges):
            verts[i] = tuple(edge.Vertexes[0].Point.multiply(scaling))
            cartPnt = ifcfile.createIfcCartesianPoint(verts[i])
            vertPnt = ifcfile.createIfcVertexPoint(cartPnt)
        orientedEdges = [None for _ in range(len(edges))]
        for i, vert in enumerate(verts):
            v2Index = (i + 1) if i < len(verts) - 1 else 0
            cartPnt1 = ifcfile.createIfcCartesianPoint(vert)
            cartPnt2 = ifcfile.createIfcCartesianPoint(verts[v2Index])
            vertPnt1 = ifcfile.createIfcVertexPoint(cartPnt1)
            vertPnt2 = ifcfile.createIfcVertexPoint(cartPnt2)
            edge = ifcfile.createIfcEdge(vertPnt1, vertPnt2)
            orientedEdges[i] = ifcfile.createIfcOrientedEdge(None, None, edge, True)
        edgeLoop = ifcfile.createIfcEdgeLoop(tuple(orientedEdges))
        # sets local coordinate system
        localPlacement = ifcbin.createIfcLocalPlacement()
        # sets face origin to the first vertex point of the planar surface
        origin = cartPnt2
        # find crossVect that is perpendicular to the planar surface
        vect0 = FreeCAD.Vector(verts[0])
        vect1 = FreeCAD.Vector(verts[1])
        vectn = FreeCAD.Vector(verts[-1])
        vect01 = vect1.sub(vect0)
        vect0n = vectn.sub(vect0)
        crossVect = vect01.cross(vect0n)
        # normalize crossVect
        normVect = crossVect.normalize()
        xAxis = ifcfile.createIfcDirection(tuple([vect01.x, vect01.y, vect01.z]))
        zAxis = ifcfile.createIfcDirection(tuple([normVect.x, normVect.y, normVect.z]))
        localAxes = ifcfile.createIfcAxis2Placement3D(origin, zAxis, xAxis)
        plane = ifcfile.createIfcPlane(localAxes)
        faceBound = ifcfile.createIfcFaceBound(edgeLoop, True)
        face = ifcfile.createIfcFaceSurface((faceBound,), plane, True)
        topologyRep = ifcfile.createIfcTopologyRepresentation(
            structContext, "Analysis", "Face", (face,)
        )
        prodDefShape = ifcfile.createIfcProductDefinitionShape(None, None, (topologyRep,))
        # sets surface thickness
        # TODO: ATM limitations
        # - for vertical slabs (walls) or inclined slabs (ramps) the thickness is taken from the Height property
        thickness = float(obj.Height) * scaling
        # creates structural member
        structuralMember = ifcfile.createIfcStructuralSurfaceMember(
            uid(),
            ownerHistory,
            obj.Label,
            None,
            None,
            localPlacement,
            prodDefShape,
            "SHELL",
            thickness,
        )

    # check for existing connection nodes
    for vert in verts:
        vertCoord = tuple(vert)
        if vertCoord in structural_nodes:
            if structural_nodes[vertCoord]:
                # there is already another member using this point
                structPntConn = structural_nodes[vertCoord]
            else:
                # there is another member with same point, create a new node connection
                structPntConn = createStructuralNode(ifcfile, ifcbin, vert)
                structural_nodes[vertCoord] = structPntConn
            ifcfile.createIfcRelConnectsStructuralMember(
                uid(),
                ownerHistory,
                None,
                None,
                structuralMember,
                structPntConn,
                None,
                None,
                None,
                None,
            )
        else:
            # just add the point, no other member using it yet
            structural_nodes[vertCoord] = None

    # check for existing connection curves
    for edge in edges:
        verts12 = tuple(
            [
                edge.Vertexes[0].Point.x,
                edge.Vertexes[0].Point.y,
                edge.Vertexes[0].Point.z,
                edge.Vertexes[-1].Point.x,
                edge.Vertexes[-1].Point.y,
                edge.Vertexes[-1].Point.z,
            ]
        )
        verts21 = tuple(
            [
                edge.Vertexes[-1].Point.x,
                edge.Vertexes[-1].Point.y,
                edge.Vertexes[-1].Point.z,
                edge.Vertexes[0].Point.x,
                edge.Vertexes[0].Point.y,
                edge.Vertexes[0].Point.z,
            ]
        )
        verts12_in_curves = verts12 in structural_curves
        verts21_in_curves = verts21 in structural_curves
        if verts21_in_curves:
            verts = verts21
        else:
            verts = verts12
        if verts12_in_curves or verts21_in_curves:
            if structural_curves[verts]:
                # there is already another member using this curve
                strucCrvConn = structural_curves[verts]
            else:
                # there is another member with same edge, create a new curve connection
                strucCrvConn = createStructuralCurve(ifcfile, ifcbin, edge)
                structural_curves[verts] = strucCrvConn
            ifcfile.createIfcRelConnectsStructuralMember(
                uid(), None, None, None, structuralMember, strucCrvConn, None, None, None, None
            )
        else:
            # just add the curve, no other member using it yet
            structural_curves[verts] = None
    return structuralMember


def createStructuralGroup(ifcfile):
    """Assigns all structural objects found in the file to the structural model"""

    import ifcopenshell

    uid = ifcopenshell.guid.new
    ownerHistory = ifcfile.by_type("IfcOwnerHistory")[0]
    structSrfMember = ifcfile.by_type("IfcStructuralSurfaceMember")
    structCrvMember = ifcfile.by_type("IfcStructuralCurveMember")
    structPntConn = ifcfile.by_type("IfcStructuralPointConnection")
    structCrvConn = ifcfile.by_type("IfcStructuralCurveConnection")
    structModel = ifcfile.by_type("IfcStructuralAnalysisModel")[0]
    if structModel:
        members = structSrfMember + structCrvMember + structPntConn + structCrvConn
        if members:
            ifcfile.createIfcRelAssignsToGroup(
                uid(), ownerHistory, None, None, members, "PRODUCT", structModel
            )


def associates(ifcfile, aobj, sobj):
    """Associates an arch object with a struct object"""

    # This is probably not the right way to do this, ie. relate a structural
    # object with an IfcProduct. Needs to investigate more....

    import ifcopenshell

    uid = ifcopenshell.guid.new
    ownerHistory = ifcfile.by_type("IfcOwnerHistory")[0]
    ifcfile.createIfcRelAssignsToProduct(uid(), ownerHistory, None, None, [sobj], None, aobj)



# ***************************************************************************
# *   Copyright (c) 2017 Markus Hovorka <m.hovorka@live.de>                 *
# *   Copyright (c) 2022 Uwe Stöhr <uwestoehr@lyx.org>                      *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD FEM solver Elmer equation object Flow"
__author__ = "Markus Hovorka, Uwe Stöhr"
__url__ = "https://www.freecad.org"

## \addtogroup FEM
#  @{

from femtools import femutils
from . import nonlinear
from ... import equationbase

CONVECTION_TYPE = ["None", "Computed", "Constant"]
FLOW_MODEL = ["Full", "No convection", "Stokes"]


def create(doc, name="Flow"):
    return femutils.createObject(doc, name, Proxy, ViewProxy)


class Proxy(nonlinear.Proxy, equationbase.FlowProxy):

    Type = "Fem::EquationElmerFlow"

    def __init__(self, obj):
        super().__init__(obj)

        obj.addProperty(
            "App::PropertyBool",
            "DivDiscretization",
            "Flow",
            (
                "Set to true for incompressible flow for more stable\n"
                "discretization when Reynolds number increases"
            ),
            locked=True,
        )
        obj.addProperty(
            "App::PropertyEnumeration", "FlowModel", "Flow", "Flow model to be used", locked=True
        )
        obj.addProperty(
            "App::PropertyBool",
            "GradpDiscretization",
            "Flow",
            (
                "If true pressure Dirichlet boundary conditions can be used.\n"
                "Also mass flux is available as a natural boundary condition."
            ),
            locked=True,
        )
        obj.addProperty(
            "App::PropertyString",
            "Variable",
            "Flow",
            "Only for a 2D model change the '3' to '2'",
            locked=True,
        )

        obj.addProperty(
            "App::PropertyEnumeration",
            "Convection",
            "Equation",
            "Type of convection to be used",
            locked=True,
        )
        obj.addProperty(
            "App::PropertyBool",
            "MagneticInduction",
            "Equation",
            (
                "Magnetic induction equation will be solved\n"
                "along with the Navier-Stokes equations"
            ),
            locked=True,
        )
        obj.FlowModel = FLOW_MODEL
        obj.FlowModel = "Full"
        obj.Variable = "Flow Solution[Velocity:3 Pressure:1]"
        obj.Convection = CONVECTION_TYPE
        obj.Convection = "Computed"
        obj.Priority = 10


class ViewProxy(nonlinear.ViewProxy, equationbase.FlowViewProxy):
    pass


##  @}
